@@ -5,6 +5,7 @@ use vars qw($VERSION $AUTOLOAD);
 use File::Path ();
 use IO::File ();
 use Cwd ();
+use Config;
 
 #this is stuff ripped out of mod_perl's Makefile.PL
 #there's still commented out crap
@@ -138,6 +139,19 @@ sub httpd_version {
 
 my $Is_Win32 = ($^O eq "MSWin32");
 
+sub typemaps {
+    my $typemaps = [];
+    
+    for (@INC) {
+	my $file;
+	if (-e ($file = "$_/auto/Apache/typemap")) {
+	    push @$typemaps, $file;
+	    last;
+	}
+    }
+    return $typemaps;
+}
+
 sub inc {
     my $self = shift;
     my $src  = $self->dir;
@@ -150,6 +164,14 @@ sub inc {
     return "@inc";
 }
 
+sub define {
+    my $self = shift;
+    if($Config{usethreads}) {
+	return "-DPERL_THREADS";
+    }
+    return "";
+}
+
 =pod
 
 my $src = Apache::src->new;
